#include <getopt.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "drivers/sx1239.h"
#include "drivers/sx1272.h"
#include "utils/logger.h"

#include "loragw_fpga.h"
#include "loragw_hal.h"
#include "loragw_radio.h"
#include "loragw_reg.h"
#include "loragw_spi.h"
#include "loragw_sx1272_fsk.h"
#include "parson.h"

#define FMIN                    850.0 // Datasheet says 860
#define FSTART                  863.0
#define FSTOP                   870.0
#define FMAX                    1020.0

#define FSTEP_MIN               0.0005
#define FSTEP                   0.2

#define FILENAME                "scan.csv"

#define SX1239_DEFAULT_RSSI_OFFSET    -13
#define SX1272_DEFAULT_RSSI_OFFSET    -7
#define RSSI_OFFSET_UNDEFINED         1234

#define DEFAULT_RSSI_OFFSET           RSSI_OFFSET_UNDEFINED
#define DEFAULT_BANDWIDTH             125000
#define DEFAULT_BITRATE               1000

#define HARDWARE_SX1272         "sx1272"
#define HARDWARE_SX1239         "sx1239"

enum hardware {
	HW_SX1239,
	HW_SX1272
};

#define SX1239_SPIDEV           "/dev/spidev1.0"
#define SX1239_GPIOEN           10
#define SX1239_GPIORST          135
#define SX1239_BITRATE          600
#define SX1239_LNA_GAIN         1
#define SX1239_DCC_FREQ         1

struct sx1239_scan_parameters {
	uint32_t fstart;
	uint32_t fstop;
	uint32_t fstep;
	uint16_t samples;
	uint32_t bandwidth;
	uint32_t bitrate;
	uint8_t lna_gain;
	uint8_t dcc_freq;
	void (*progress)(uint32_t, uint32_t);
};

struct sx1272_scan_parameters {
	char *conf_filename;
	uint32_t fstart;
	uint32_t fstop;
	uint32_t fstep;
	uint16_t samples;
	uint32_t bandwidth;
	uint32_t bitrate;
	uint8_t lna_gain;
};

static bool running = true;
static bool graceful_exit = true;
static void *sx1239 = NULL;

static void signal_handler(int signal)
{
	switch (signal) {
	case SIGQUIT:
		graceful_exit = false;
		/* Fallthrough */
	case SIGINT:
		/* Fallthrough */
	case SIGTERM:
		running = false;
		if (sx1239) {
			sx1239_abort(sx1239);
		}
		if (lgw_isstarted()) {
			lgw_stop();
		}
		break;
	}
}

static int handle_signals(void)
{
	struct sigaction signal_action;

	signal_action.sa_handler = signal_handler;
	sigemptyset(&signal_action.sa_mask);
	signal_action.sa_flags = 0;

	if (sigaction(SIGINT, &signal_action, NULL) < 0) {
		eerror("Failed to handle SIGINT");
		return -1;

	} else if (sigaction(SIGQUIT, &signal_action, NULL) < 0) {
		eerror("Failed to handle SIGQUIT");
		return -1;

	} else if (sigaction(SIGTERM, &signal_action, NULL) < 0) {
		eerror("Failed to handle SIGTERM");
		return -1;
	}

	return 0;
}

static void print_histogram_freq(uint16_t *histogram, uint32_t freq,
                                 uint16_t samples, int16_t offset, uint8_t basic)
{
	uint16_t h, s;
	float r, a;
	int32_t i;

	for (i = SX1239_HISTO_VALUE_MAX, a = 0.0, s = 0; i >= 0; i--) {
		r = offset - i / 2.0;
		h = histogram[i];

		if (h != 0) {
			a = (a * s + r * h) / (s + h);
			s += h;
		}
	}

	if (basic == 0) {
		fprintf(stdout, "%10u Hz %-6.1f ", freq, a);

	} else {
		fprintf(stdout, "%u %.1f", freq, a);
	}

	for (i = SX1239_HISTO_VALUE_MAX; i >= 0 && basic == 0; i -= 4) {
		h = histogram[i - 0] + histogram[i - 1] +
		    histogram[i - 2] + histogram[i - 3];

		if (h == 0) {
			putc(' ', stdout);

		} else if (h < samples / 100) {
			putc('.', stdout);

		} else {
			putc('|', stdout);
		}
	}

	putc('\n', stdout);
}

static void print_histogram(uint16_t *histogram,
                            uint32_t fstart, uint32_t fstop, uint32_t fstep,
                            uint16_t samples, int16_t offset, uint8_t basic)
{
	uint16_t *h;
	uint32_t f;

	if (basic == 0) {
		fprintf(stdout, "\n%-20s | %6.1f dBm %45.1f dBm |\n",
		        "Histogram scale",
		        (float)offset - UINT8_MAX / 2.0,
		        (float)offset);
	}

	for (h = histogram, f = fstart;
	     f <= fstop;
	     f += fstep, h += SX1239_HISTO_STEP_NB) {
		print_histogram_freq(h, f, samples, offset, basic);
	}
}

static int save_histogram_freq(FILE *file, uint16_t *histogram, uint32_t freq,
                               int16_t offset, uint8_t strip_zero)
{
	int32_t i;

	if (fprintf(file, "%u", freq) < 0) {
		error("Failed to print histogram header");
		return -1;
	}

	for (i = SX1239_HISTO_VALUE_MAX; i >= 0; i--) {
		if (strip_zero == 0 || histogram[i] != 0) {
			if (fprintf(file, ",%.1f,%u",
			            offset - i / 2.0, histogram[i]) < 0) {
				error("Failed to print histogram value");
				return -1;
			}
		}
	}

	if (fprintf(file, "\n") < 0) {
		error("Failed to print histogram footer");
		return -1;
	}

	return 0;
}

static int save_histogram(FILE *file, uint16_t *histogram,
                          uint32_t fstart, uint32_t fstop, uint32_t fstep,
                          int16_t offset, uint8_t strip_zero)
{
	uint16_t *h;
	uint32_t f;

	for (h = histogram, f = fstart;
	     f <= fstop;
	     f += fstep, h += SX1239_HISTO_STEP_NB) {
		if (save_histogram_freq(file, h, f, offset, strip_zero) < 0) {
			return -1;
		}
	}

	return 0;
}

static void print_progress(uint32_t part, uint32_t total)
{
	if (total == 0) {
		fprintf(stdout, "\r%7s\r", " ");
		fflush(stdout);

	} else {
		fprintf(stdout, "\r%5.1f %%", 100.0 * part / total);
		fflush(stdout);
	}
}

uint16_t* sx1239_scan(struct sx1239_scan_parameters* scan_params)
{
	uint16_t *histogram = NULL;

	/* Initialize sniffer */
	if ((sx1239 = sx1239_open(SX1239_SPIDEV,
	                          SX1239_GPIOEN,
	                          SX1239_GPIORST)) == NULL) {
		error("Failed to open SX1239");
		goto sx1239_scan_exit;

	} else if (sx1239_set_bw(sx1239, scan_params->bandwidth) < 0) {
		error("Failed to set SX1239 bandwidth");
		goto sx1239_scan_exit;

	} else if (sx1239_set_br(sx1239, scan_params->bitrate) < 0) {
		error("Failed to set SX1239 bit rate");
		goto sx1239_scan_exit;

	} else if (sx1239_set_lna_gain(sx1239, scan_params->lna_gain) < 0) {
		error("Failed to set SX1239 LNA gain");
		goto sx1239_scan_exit;

	} else if (sx1239_set_dcc_freq(sx1239, scan_params->dcc_freq) < 0) {
		error("Failed to set SX1239 DCC cutoff frequency");
		goto sx1239_scan_exit;
	}

	info("Bandwidth:       %17u Hz", scan_params->bandwidth);
	info("Bit rate:        %16u bps", scan_params->bitrate);
	info("LNA gain:        %19s%1u",
	     "G", (scan_params->lna_gain > 6) ? 1 : (scan_params->lna_gain < 1) ? 1 : scan_params->lna_gain);
	info("DCC freq:        %18.3g %%",
	     (1 << ((scan_params->dcc_freq > 8) ? 0 : (scan_params->dcc_freq < 1) ? 0 : scan_params->dcc_freq - 1)) / 8.0);

	if ((histogram = sx1239_oneshot_histogram(sx1239,
	                     scan_params->fstart, scan_params->fstop,
	                     scan_params->fstep, scan_params->samples,
	                     scan_params->progress)) == NULL) {
		error("Failed to get SX1239 histogram");
		goto sx1239_scan_exit;
	}
sx1239_scan_exit:
	return histogram;
}

uint16_t* sx1272_scan(struct sx1272_scan_parameters* scan_params)
{
	uint16_t *histogram = NULL;
	int x;

	if (lgw_connect(true, 0) < 0) {
		goto sx1272_scan_exit;
	}

	x = lgw_setup_sx127x(scan_params->fstart, MOD_FSK, LGW_SX127X_RXBW_62K5_HZ, 0);
	if (x != 0) {
		error("Something failed in SX1272 configuration");
		goto sx1272_scan_exit;
	}

	if ( (scan_params->bandwidth != DEFAULT_BANDWIDTH) ||
	     (scan_params->bitrate != DEFAULT_BITRATE) ) {

		/* go to standby mode */
		if(sx1272_set_mode(SX1272_REG_OPMODE_MODE_STANDBY) != 0) {
			error("mode standby KO");
			goto sx1272_scan_exit;
		} else {
			info("Setting mode standby OK");
		}

		/* bandwidth */
		if(sx1272_set_bw(scan_params->bandwidth) != 0)
			goto sx1272_scan_exit;

		/* bitrate in spectral scan */
		if(sx1272_set_br(scan_params->bitrate) != 0)
			goto sx1272_scan_exit;

		/* RX mode */
		if(sx1272_set_mode(SX1272_REG_OPMODE_MODE_RX) != 0) {
			error("Failed to set mode RX");
			sx1272_dumpRegs();
			goto sx1272_scan_exit;
		}
		info("Setting mode RX OK");
		//sx1272_dumpRegs();
	}

	histogram = sx1272_histogram(scan_params->fstart, scan_params->fstop,
	                             scan_params->fstep, scan_params->samples);

sx1272_scan_exit:
	//lgw_disconnect();
	lgw_stop();
	return histogram;
}


int main(int argc, char **argv)
{
	struct timespec start, end;
	int status = EXIT_FAILURE;
	int opt, iopt, itm;

	char *filename = NULL;
	FILE *file = NULL;

	sniffer_type_e hw = SNIFFER_SX1272;

	uint32_t fstart = FSTART * 1E6;
	uint32_t fstop = FSTOP * 1E6;
	uint32_t fstep = FSTEP * 1E6;
	double fvector[3];

	uint16_t samples = 1024;
	int16_t offset = DEFAULT_RSSI_OFFSET;
	uint16_t *histogram = NULL;

	uint32_t bandwidth = 0;
	uint32_t bitrate = SX1239_BITRATE;
	uint8_t lna_gain = SX1239_LNA_GAIN;
	uint8_t dcc_freq = SX1239_DCC_FREQ;

	uint8_t start_lna = 0;
	uint8_t strip_zero = 0;
	uint8_t basic = 0;
	uint8_t force = 0;

	void (*progress)(uint32_t, uint32_t) = &print_progress;

	struct sx1239_scan_parameters sx1239_params;
	struct sx1272_scan_parameters sx1272_params;

	eeprom_config_t config;

	const struct option opts[] = {
		{"hardware",   required_argument,  NULL, 'H'},
		{"frequency",  required_argument,  NULL, 'f'},
		{"samples",    required_argument,  NULL, 's'},
		{"output",     required_argument,  NULL, 'o'},
		{"offset",     required_argument,  NULL, 'O'},
		{"bandwidth",  required_argument,  NULL, 'b'},
		{"bitrate",    required_argument,  NULL, 'B'},
		{"lna-gain",   required_argument,  NULL, 'g'},
		{"dcc-freq",   required_argument,  NULL, 'd'},
		{"start-lna",  no_argument,        NULL, 'l'},
		{"strip-zero", no_argument,        NULL, '0'},
		{"basic",      no_argument,        NULL,  0 },
		{"force",      no_argument,        NULL,  0 },
		{"version",    no_argument,        NULL, 'v'},
		{"help",       no_argument,        NULL, 'h'},
		{NULL,         0,                  NULL,  0 },
	};

	/* Get start time */
	if (clock_gettime(CLOCK_MONOTONIC, &start) < 0) {
		error("Failed to get start time");
		goto exit;
	}

	/* Manage options */
	while ((opt = getopt_long(argc, argv,
	                          "H:c:f:s:o:O:b:B:g:d:l0vh", opts, &iopt)) != EOF) {
		switch (opt) {
		case 'H':
			if (strcmp(optarg, HARDWARE_SX1272) == 0) {
				hw = SNIFFER_SX1272;
			} else if (strcmp(optarg, HARDWARE_SX1239) == 0) {
				hw = SNIFFER_SX1239;
			} else {
				error("Failed to parse hardware");
				goto exit;
			}
			break;

		case 'f':
			if ((itm = sscanf(optarg, "%lf:%lf:%lf", &fvector[0],
			                  &fvector[1], &fvector[2])) < 0) {
				error("Failed to parse frequency vector");
				goto exit;

			}

			switch (itm) {
			case 1:
				fvector[2] = fvector[0];
				fvector[1] = FSTEP;
				break;

			case 2:
				fvector[2] = fvector[1];
				fvector[1] = FSTEP;
				break;
			}

			if (fvector[0] > fvector[2]) {
				error("FSTART is greater than FSTOP");
				goto exit;

			} else if (fvector[0] < FMIN && force == 0) {
				error("FSTART is lower than %f", FMIN);
				goto exit;

			} else if (fvector[2] > FMAX && force == 0) {
				error("FSTOP is greater than %f", FMAX);
				goto exit;

			} else if (fvector[1] < FSTEP_MIN && force == 0) {
				error("FSTEP is lower than %f", FSTEP_MIN);
				goto exit;
			}

			fstart = fvector[0] * 1E6;
			fstep = fvector[1] * 1E6;
			fstop = fvector[2] * 1E6;
			break;

		case 's':
			if (sscanf(optarg, "%hu", &samples) < 1) {
				eerror("Failed to read samples");
				goto exit;

			} else if (samples == 0) {
				error("Samples number cannot be null");
				goto exit;
			}

			break;

		case 'o':
			free(filename);

			if ((filename = strdup(optarg)) == NULL) {
				eerror("Failed to duplicate filename");
				goto exit;
			}

			break;

		case 'O':
			if (sscanf(optarg, "%hi", &offset) < 1) {
				eerror("Failed to read offset");
				goto exit;
			}

			break;

		case 'b':
			if (sscanf(optarg, "%u", &bandwidth) < 1) {
				eerror("Failed to read bandwidth");
				goto exit;
			}

			break;

		case 'B':
			if (sscanf(optarg, "%u", &bitrate) < 1) {
				eerror("Failed to read bitrate");
				goto exit;
			}

			break;

		case 'g':
			if (sscanf(optarg, "%hhu", &lna_gain) < 1) {
				eerror("Failed to read LNA gain");
				goto exit;
			}

			break;

		case 'd':
			if (sscanf(optarg, "%hhu", &dcc_freq) < 1) {
				eerror("Failed to read DCC cutoff frequency");
				goto exit;
			}

			break;

		case 'l':
			start_lna = 1;
			break;

		case '0':
			strip_zero = 1;
			break;

		case 0:
			if (strcmp("basic", opts[iopt].name) == 0) {
				progress = NULL;
				basic = 1;

			} else if (strcmp("force", opts[iopt].name) == 0) {
				force = 1;
			}

			break;

		case 'v':
			fprintf(stdout, "klk_spectral_scan " PACKAGE_VERSION "\n");
			status = EXIT_SUCCESS;
			goto exit;

		default:
			fprintf(stderr, "Usage: klk_spectral_scan [OPTIONS]...\n"
			        "\n"
			        "-H, --hardware sx1239,sx1272. Defaulted to sx1239\n"
			        "-c, --conf FILENAME"
			        "    Configuration file (for sx1272 case)\n"
			        "-f, --frequency FSTART:FSTEP:FSTOP"
			        "    Scan frequency vector (in MHz).\n"
			        "-s, --samples NUMBER              "
			        "    Number of RSSI samples per scan.\n"
			        "-o, --output FILENAME             "
			        "    Output filename.\n"
			        "-O, --offset OFFSET               "
			        "    RSSI offset (in dBm).\n"
			        "-b, --bandwidth BANDWIDTH         "
			        "    Use the specified bandwidth (in Hz).\n"
			        "-B, --bitrate BITRATE             "
			        "    Use the specified bit rate (in bps).\n"
			        "-g, --lna-gain LNA_GAIN           "
			        "    Use the specified LNA gain.\n"
			        "                                  "
			        "    From 1 (G1) to 6 (G6). Defaulted to 1.\n"
			        "-d, --dcc-freq DCC_FREQ           "
			        "    Use the specified DCC cutoff frequency.\n"
			        "                                  "
			        "    From 1 (0.125) to 8 (16). Defaulted to 1.\n"
			        "-l, --start-lna                   "
			        "    Start LNA if not already done.\n"
			        "-0, --strip-zero                  "
			        "    Don't save zero values.\n"
			        "    --basic                       "
			        "    Print script friendly outputs.\n"
			        "-v, --version                     "
			        "    Display version.\n"
			        "-h, --help                        "
			        "    Display this help screen.\n");
			goto exit;
		}
	}

	if (bandwidth == 0) {
		bandwidth = fstep;
	}

	if (start_lna == 1 && lna_start(fstart) < 0) {
		error("Failed to start LNA");
		goto exit;
	}

	/* Handle signals */
	if (handle_signals() < 0) {
		error("Failed to handle signals");
		goto exit;
	}

	/* Open output file */
	if (filename == NULL && (filename = strdup(FILENAME)) == NULL) {
		eerror("Failed to duplicate filename");
		goto exit;

	} else if ((file = fopen(filename, "w")) == NULL) {
		eerror("Failed to open output file");
		goto exit;
	}

	/* Check Offset */
	if(offset == RSSI_OFFSET_UNDEFINED)
	{
		/* Try to get it from EEPROM if present */
		if (lgw_connect(true, LGW_DEFAULT_NOTCH_FREQ) < 0) {
			eerror("Failed to connect FPGA");
			goto exit;
		} else if (lgw_eeprom_readbuf(0, config.raw, sizeof(config.raw)) < 0) {
			eerror("Failed to read EEPROM");
			lgw_disconnect();
			goto exit;
		}
		lgw_disconnect();

		hw = lgw_eeprom_get_sniffer_type(config.s.type);
		if ( (config.s.manuf == MANUFACTURER_KERLINK) && (hw == SNIFFER_SX1272) )
		{
			/* Get calibrated offset for iStation/iFevo/etc... */
			info("Use calibrated RSSI offset for SX1272 sniffer");
			offset = config.s.sx1272_rssi_offset;
		}
		else {
			info("Use default RSSI offset value");
			if (hw == SNIFFER_SX1239) offset = SX1239_DEFAULT_RSSI_OFFSET;
			else offset = SX1272_DEFAULT_RSSI_OFFSET;
		}
	}

	info("Start frequency: %17u Hz", fstart);
	info("Stop frequency:  %17u Hz", fstop);
	info("Step frequency:  %17u Hz", fstep);
	info("RSSI samples:    %20u", samples);
	info("Output filename: %20s", filename);
	info("RSSI offset:     %16i dBm", offset);

	if (hw == SNIFFER_SX1239) {
		sx1239_params.fstart = fstart;
		sx1239_params.fstop = fstop;
		sx1239_params.fstep = fstep;
		sx1239_params.samples = samples;
		sx1239_params.bandwidth = bandwidth;
		sx1239_params.bitrate = bitrate;
		sx1239_params.lna_gain = lna_gain;
		sx1239_params.dcc_freq = dcc_freq;
		sx1239_params.progress = progress;
		histogram = sx1239_scan(&sx1239_params);
	} else if (hw == SNIFFER_SX1272) {
		sx1272_params.fstart = fstart;
		sx1272_params.fstop = fstop;
		sx1272_params.fstep = fstep;
		sx1272_params.samples = samples;
		sx1272_params.bandwidth = bandwidth;
		sx1272_params.bitrate = bitrate;
		sx1272_params.lna_gain = lna_gain;
		histogram = sx1272_scan(&sx1272_params);
	} else {
		error("Unknown hw type");
		goto exit;
	}

	if (histogram == NULL)
		goto exit;

	if (save_histogram(file, histogram, fstart, fstop,
	                          fstep, offset, strip_zero) < 0) {
		error("Failed to save histogram");
		goto exit;

	} else {
		print_histogram(histogram, fstart, fstop, fstep,
		                samples, offset, basic);
	}

	/* Display footer message */
	if (basic == 0) {
		if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
			info("Scan %s", running ? "finished" : "exited");

		} else {
			if (end.tv_nsec < start.tv_nsec) {
				end.tv_sec -= start.tv_sec + 1;
				end.tv_nsec -= start.tv_nsec - 1E9;

			} else {
				end.tv_sec -= start.tv_sec;
				end.tv_nsec -= start.tv_nsec;
			}

			info("Scan %s in %.3f second%s",
			     running ? "finished" : "exited",
			     end.tv_sec + end.tv_nsec / 1E9,
			     end.tv_sec ? "s" : "");
		}
	}

	status = graceful_exit ? EXIT_SUCCESS : EXIT_FAILURE;

exit:   /* Clean up */

	if (sx1239 != NULL) {
		sx1239_close(sx1239);
	}

	if (file != NULL) {
		fclose(file);
	}

	if (start_lna == 1) {
		lna_stop();
	}

	if (histogram)
		free(histogram);
	free(filename);
	exit(status);
}
