/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include "drivers/sx1272.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "loragw_radio.h"
#include "loragw_spi.h"
#include "loragw_sx1272_fsk.h"
#include "utils/logger.h"
#include "utils/time.h"


#define FXOSC                           32000000

/**
@struct lgw_radio_FSK_bandwidth_s
@brief Associate a bandwidth in kHz with its corresponding register values
*/
struct lgw_sx1272_FSK_bandwidth_s {
    uint32_t    RxBwKHz;
    uint8_t     RxBwMant;
    uint8_t     RxBwExp;
};

const struct lgw_sx1272_FSK_bandwidth_s sx1272_FskBandwidths[] =
{
    { 2600  , 2, 7 },   /* LGW_SX127X_RXBW_2K6_HZ */
    { 3100  , 1, 7 },   /* LGW_SX127X_RXBW_3K1_HZ */
    { 3900  , 0, 7 },   /* ... */
    { 5200  , 2, 6 },
    { 6300  , 1, 6 },
    { 7800  , 0, 6 },
    { 10400 , 2, 5 },
    { 12500 , 1, 5 },
    { 15600 , 0, 5 },
    { 20800 , 2, 4 },
    { 25000 , 1, 4 },   /* ... */
    { 31300 , 0, 4 },
    { 41700 , 2, 3 },
    { 50000 , 1, 3 },
    { 62500 , 0, 3 },
    { 83333 , 2, 2 },
    { 100000, 1, 2 },
    { 125000, 0, 2 },
    { 166700, 2, 1 },
    { 200000, 1, 1 },   /* ... */
    { 250000, 0, 1 }    /* LGW_SX127X_RXBW_250K_HZ */
};

int sx1272_reg_check(uint8_t addr, uint8_t val, uint8_t mask)
{
	int count;
	uint8_t reg_val;

	for (count = 0; count < 100; count++) {
		if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(addr, &reg_val)) {
			error("Failed to get register 0x%x", addr);
			return -1;
		} else if ((reg_val & mask) == val)
			return 0;
		else
			sleep_us(1000);
	}

	error("Check failed for register 0x%x: 0x%x != 0x%x", addr, reg_val, val);
	return -1;
}

int sx1272_reg_set_mask(uint8_t addr, uint8_t val, uint8_t mask)
{
	uint8_t reg_val;
	uint8_t new_val;

	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(addr, &reg_val)) {
		error("Failed to get register");
		return -1;
	}

	new_val = reg_val & ~mask;
	new_val |= val & mask;

	if (new_val != reg_val) {
		if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(addr, new_val)) {
			error("Failed to set register");
			return -1;
		}
	}
	return 0;
}

int sx1272_set_mode(uint8_t mode)
{
	uint8_t reg_val;

	if (sx1272_reg_set_mask(SX1272_REG_OPMODE, mode, SX1272_REG_OPMODE_MODE_MASK)) {
		error("Failed to set opmode register");
		return -1;
	}
	if (sx1272_reg_check(SX1272_REG_OPMODE, mode,
	                     SX1272_REG_OPMODE_MODE_MASK) < 0) {
				lgw_sx127x_reg_r(SX1272_REG_OPMODE, &reg_val);
				error("Mode in OPMODE reg is 0x%02x instead of 0x%02x", reg_val, mode);
				return -1;
	}
	lgw_sx127x_reg_r(SX1272_REG_OPMODE, &reg_val);
	/*info("sx1272_set_mode OPMODE value: 0x%x", reg_val);*/

	if (sx1272_reg_check(SX1272_REG_IRQFLAGS1, SX1272_REG_IRQFLAGS1_MODE_READY,
	                     SX1272_REG_IRQFLAGS1_MODE_READY) < 0) {
		error("Programmed mode not correctly set");
		return -1;
	}
	return 0;
}

int sx1272_set_br(uint32_t br)
{
	uint16_t bitrate;

	bitrate = (FXOSC / br);

	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(SX1272_REG_BITRATEMSB, (bitrate & 0xFF00) >> 8)) {
		error("Failed to set BITRATE MSB");
		return -1;
	}
	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(SX1272_REG_BITRATELSB, (bitrate & 0x00FF))) {
		error("Failed to set BITRATE LSB");
		return -1;
	}

	return 0;
}

int sx1272_set_bw(uint32_t bw)
{
	uint8_t rxBwExp;
	uint8_t rxBwMant;
	unsigned int i;

	/* we need to use mid bandwidth in KHz */
	for (i=0; i< sizeof(sx1272_FskBandwidths) /
			sizeof(struct lgw_sx1272_FSK_bandwidth_s); i++) {
		if (bw <= sx1272_FskBandwidths[i].RxBwKHz * 2) {
			rxBwExp = sx1272_FskBandwidths[i].RxBwExp;
			rxBwMant = sx1272_FskBandwidths[i].RxBwMant;
			if (LGW_SPI_SUCCESS != lgw_sx127x_reg_w(SX1272_REG_RXBW,
						rxBwExp | (rxBwMant << 3))) {
				error("Failed to set bandwidth");
				return -1;
			}
			info("Selected single side bandwidth: %lu", (unsigned long) sx1272_FskBandwidths[i].RxBwKHz);
			return 0;
		}
	}

	error("bandwidth (%lu) not supported", (unsigned long)bw);
	return -1;
}

int sx1272_set_freq(uint32_t freq)
{
	uint64_t freq_reg;
	int x;

	freq_reg = ((uint64_t)freq << 19) / (uint64_t)32000000;
	x = lgw_sx127x_reg_w(SX1272_REG_FRFMSB, (freq_reg >> 16) & 0xFF);
	x = lgw_sx127x_reg_w(SX1272_REG_FRFMID, (freq_reg >> 8) & 0xFF);
	x = lgw_sx127x_reg_w(SX1272_REG_FRFLSB, (freq_reg >> 0) & 0xFF);
	if (LGW_SPI_SUCCESS != x) {
		error("Failed to set frequency");
		return -1;
	}

	return 0;
}

/* rssi in db is 0 - result /2 */
int sx1272_rssi(uint8_t firstAtThisFreq, int* rssi)
{
	uint8_t reg_rssi;

	/* start measure */
	if (firstAtThisFreq) {
		if (sx1272_reg_set_mask(SX1272_REG_RXCONFIG, 
		       SX1272_REG_RXCONFIG_RESTART_WITH_PLL_LOCK_MASK, 
		       SX1272_REG_RXCONFIG_RESTART_WITH_PLL_LOCK_MASK)) {
			error("Failed to set RXCONFIG register (with pll)");
			return -1;
		}
	} else {
		if (sx1272_reg_set_mask(SX1272_REG_RXCONFIG, 
		       SX1272_REG_RXCONFIG_RESTART_WITHOUT_PLL_LOCK_MASK, 
		       SX1272_REG_RXCONFIG_RESTART_WITHOUT_PLL_LOCK_MASK)) {
			error("Failed to set RXCONFIG register (without pll)");
			return -1;
		}
	}

	/* wait rx ready flag */
	if (sx1272_reg_check(SX1272_REG_IRQFLAGS1, SX1272_REG_IRQFLAGS1_RX_READY,
	                  SX1272_REG_IRQFLAGS1_RX_READY) < 0) {
		error("SX1272 not RX ready");
		return -1;
	}

	/* read result */
	if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(SX1272_REG_RSSIVALUE, &reg_rssi)) {
		error("Failed to read RSSI VALUE register");
		return -1;
	}

	/*rssidb = 0.0f - ((float)reg_rssi) / 2;*/
	*rssi = reg_rssi;
	return 0;
}

uint16_t *sx1272_histogram(uint32_t fstart, uint32_t fstop,
                           uint32_t fstep, uint32_t samples)
{
	uint32_t steps = (fstop - fstart) / fstep + 1;
	uint32_t i, step;
	uint16_t *histogram;
	int rssi;
	double duration_set_freq = 0;
	double duration_rssi = 0;
	struct timespec temp;

	if ((histogram = calloc(steps * SX1272_HISTO_STEP_NB,
	                               sizeof(uint16_t))) == NULL) {
		error("Failed to allocate memory");
		return NULL;
	}

	for (i = 0; i < samples; i++) {
		for (step = 0; step < steps; step++) {
			clock_gettime(CLOCK_REALTIME, &temp);
			if (sx1272_set_freq(fstart + step * fstep) != 0) {
				error("Failed to set frequency");
				goto sx1272_histogram_error;
			}
			duration_set_freq += get_duration_since(&temp);
			clock_gettime(CLOCK_REALTIME, &temp);
			if (sx1272_rssi(1, &rssi) < 0) {
				error("Failed to read RSSI");
				goto sx1272_histogram_error;
			}
			duration_rssi += get_duration_since(&temp);
			/* store result in histogram */
			histogram[step * SX1272_HISTO_STEP_NB + rssi]++;
		}
	}

	/* info("duration_set_freq: %lf, duration_rssi:%lf\n",
			duration_set_freq, duration_rssi); */

	return histogram;
sx1272_histogram_error:
	free(histogram);
	return NULL;
}

void sx1272_dumpRegs(void)
{
	int i;
	uint8_t reg_val;

	info("Dumping SX1272 registers");
	for(i=0; i<0x71; i++) {
		if (LGW_SPI_SUCCESS != lgw_sx127x_reg_r(i, &reg_val)) {
			info("%02x: error", i);
		} else {
			info("%02x: %02x", i, reg_val);
		}
	}
}
