/*
  _  __                _   _           _
 | |/ /   ___   _ __  | | (_)  _ __   | | __
 | ' /   / _ \ | '__| | | | | | '_ \  | |/ /
 | . \  |  __/ | |    | | | | | | | | |   <
 |_|\_\  \___| |_|    |_| |_| |_| |_| |_|\_\

Description:
	EEPROM write tool

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Martin Chaplet <m.chaplet@kerlink.fr>
*/


/* -------------------------------------------------------------------------- */
/* --- DEPENDANCIES --------------------------------------------------------- */

/* fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
	#define _XOPEN_SOURCE 600
#else
	#define _XOPEN_SOURCE 500
#endif

#include <stdint.h>		/* C99 types */
#include <stdbool.h>	/* bool type */
#include <stdio.h>		/* printf fprintf sprintf fopen fputs */

#include <signal.h>		/* sigaction */
#include <unistd.h>		/* getopt access */
#include <getopt.h>
#include <stdlib.h>		/* rand */

#include <string.h>     /* memset */
//#include "loragw_hal.h"
#include "loragw_reg.h"
#include "loragw_fpga.h"

#if KERLINK

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

#ifndef round
#define round(x) ((x)>=0?(long)((x)+0.5):(long)((x)-0.5))
#endif

#define ARRAY_SIZE(a)	(sizeof(a) / sizeof((a)[0]))
#define DEBUG_PRINTF(fmt, args...)	fprintf(stdout,"%s:%d: "fmt, __FUNCTION__, __LINE__, args)
/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS ---------------------------------------------------- */

#ifndef VERSION
#define VERSION "Unknown"
#endif
/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES (GLOBAL) ------------------------------------------- */



/* signal handling variables */
struct sigaction sigact; /* SIGQUIT&SIGINT&SIGTERM signal handling */
static int exit_sig = 0; /* 1 -> application terminates cleanly (shut down hardware, close open files, etc) */
static int quit_sig = 0; /* 1 -> application terminates without shutting down the hardware */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

static void sig_handler(int sigio);

void usage (void);

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ----------------------------------------- */

static void sig_handler(int sigio) {
	if (sigio == SIGQUIT) {
		quit_sig = 1;
	} else if ((sigio == SIGINT) || (sigio == SIGTERM)) {
		exit_sig = 1;
	}
}

/* describe command line options */
void usage(void) {
    printf("Version: " VERSION "\n");
	printf( "Available options:\n");
	printf( " -h print this help\n");
	printf( " -d: Display EEPROM content\n");
	printf( " -m <manufacturer code>\n");
	printf( " -t <boardtype code>\n");
	printf( " -v <hardware version>\n");
	printf( " -s <serial number>\n");
	printf( " -e: Erase EEPROM\n");
	printf( " -f <freq_offset>: TCXO Frequency offset\n");
	printf( " -g <gain_index, Dig, PA, Mix, rf_power>\n");
	printf( " --r0_roffset <Radio0_rssi_offset>\n");
	printf( " --r1_roffset <Radio1_rssi_offset>\n");
	printf( " --sx1272_roffset <SX1272_rssi_offset>\n");
}

static eeprom_config_t eeconfig;

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main(int argc, char **argv)
{
    int option_index = 0;

    static struct option long_options[] = {
        {"r0_roffset", 1, 0, 0},
        {"r1_roffset", 1, 0, 0},
        {"sx1272_roffset", 1, 0, 0},
        {0, 0, 0, 0}
    };
	int i, x, ret = EXIT_SUCCESS;
	unsigned long serialno = 0;
	bool b_write = false;
	uint8_t gain_index, dig_gain, pa_gain, mix_gain;
	float rf_power, rssi_offset;
	
	memset(eeconfig.raw, 0, sizeof(eeconfig.raw));
	
	/* configure signal handling */
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	sigact.sa_handler = sig_handler;
	sigaction(SIGQUIT, &sigact, NULL);
	sigaction(SIGINT, &sigact, NULL);
	sigaction(SIGTERM, &sigact, NULL);
	
	/* start SPI link */
    x = lgw_connect(true, 0);
    if( x != 0 ) {
        printf( "ERROR: Failed to connect to FPGA\n" );
        return EXIT_FAILURE;
    }

    // (Re-)Read EEPROM config
    lgw_eeprom_readbuf(0, eeconfig.raw, sizeof(eeconfig.raw));
    if (eeconfig.s.manuf == 0xFF) {
      printf("ERROR: EEPROM seems empty\n");
      memset(eeconfig.raw, 0, sizeof(eeconfig.raw));
    }

    /* parse command line options */
    while( (i = getopt_long (argc, argv, "def:g:m:r:s:t:v:h", long_options, &option_index)) != -1 )
    {
      switch (i) {
        case 'h':
          usage();
          return EXIT_FAILURE;
          break;



        case 0:
          if(option_index < 3)
          {
            // RSSI Offset type
            i = sscanf(optarg, "%f", &rssi_offset);
            if(i != 1)
            {
              printf( "ERROR: argument parsing of --XXX_roffset argument. Use -h to print help\n" );
              return EXIT_FAILURE;
            }

            if( strcmp(long_options[option_index].name,"r0_roffset") == 0 )
            {
              eeconfig.s.radio0_rssi_offset = rssi_offset;
            }
            else if( strcmp(long_options[option_index].name,"r1_roffset") == 0 )
            {
              eeconfig.s.radio1_rssi_offset = rssi_offset;
            }
            else if( strcmp(long_options[option_index].name,"sx1272_roffset") == 0 )
            {
              eeconfig.s.sx1272_rssi_offset = rssi_offset;
            }
            b_write = true;
          }
          else
          {
            printf( "Unknown argument\n" );
            return EXIT_FAILURE;
          }
          break;

        // Erase struct
        case 'e':
          memset(eeconfig.raw, 0, sizeof(eeconfig.raw));
          b_write = true;
          break;

        // Get Freq Offset
        case 'f':
          eeconfig.s.freq_offset = strtol(optarg, NULL,10);
          b_write = true;
          break;
        // Get RSSI Offset
        case 'r':
          // Apply same config for both radios
          eeconfig.s.radio0_rssi_offset = strtof(optarg, NULL);
          eeconfig.s.radio1_rssi_offset = eeconfig.s.radio0_rssi_offset;
          b_write = true;
          break;
        // Get Gains
        case 'g':
          // Cut optarg in multiple values ("4,0,3,15,27")
          i = sscanf(optarg, "%hhu,%hhu,%hhu,%hhu,%f",
                     &gain_index, &dig_gain, &pa_gain,
                     &mix_gain, &rf_power);
          if(i != 5)
          {
            printf("Not enough/too many args for gain (must be 5)\n");
            return EXIT_FAILURE;
          }

          if(gain_index >= EEPROM_TX_LUT_MAX)
          {
            printf("Gain index must be between 0 and 15\n");
            return EXIT_FAILURE;
          }

          // Prepare structure
          eeconfig.s.tx_lut[gain_index].dig_gain = dig_gain;
          eeconfig.s.tx_lut[gain_index].pa_gain = pa_gain;
          eeconfig.s.tx_lut[gain_index].mix_gain = mix_gain;
          eeconfig.s.tx_lut[gain_index].rf_power = (int8_t) round(rf_power);
          b_write = true;
          break;


        case 'm':
          eeconfig.s.manuf = strtoul(optarg, NULL, 0);
          b_write = true;
          break;
        case 't':
          eeconfig.s.type = strtoul(optarg, NULL, 0);
          b_write = true;
          break;
        case 'v':
          eeconfig.s.hwver = strtoul(optarg, NULL, 0);
          b_write = true;
          break;
        case 's':
          serialno = strtoul(optarg, NULL, 16);
          eeconfig.s.serialno[2] = (uint8_t) (serialno & 0xFF);
          eeconfig.s.serialno[1] = (uint8_t) ((serialno >> 8) & 0xFF);
          eeconfig.s.serialno[0] = (uint8_t) ((serialno >> 16) & 0xFF);
          b_write = true;
          break;

        case 'd':
            // Just display, nothing to do
            break;

        default:
          printf("ERROR: argument parsing use -h option for help\n");
          usage();
          return EXIT_FAILURE;
      }
    }

    if(b_write)
    {
      if(LGW_REG_SUCCESS == lgw_eeprom_writebuf(0, eeconfig.raw, sizeof(eeconfig.raw)))
      {
        printf("Write OK\n");
        ret = EXIT_SUCCESS;
      }
      else
      {
        printf("ERROR writing EEPROM\n");
        ret = EXIT_FAILURE;
      }
    }

    // (Re-)Read EEPROM config
    memset(eeconfig.raw, 0, sizeof(eeconfig.raw));
    lgw_eeprom_readbuf(0, eeconfig.raw, sizeof(eeconfig.raw));

    if (eeconfig.s.manuf == 0xFF) {
        printf("ERROR: EEPROM seems empty\n");
    }
    else
    {
      printf("INFO: Lora FPGA Board detected: manuf %02X, type %u, hwver %02X, no %02X%02X%02X\n",
          eeconfig.s.manuf, eeconfig.s.type, eeconfig.s.hwver,
          eeconfig.s.serialno[0], eeconfig.s.serialno[1], eeconfig.s.serialno[2]);

      // Print out calibration information
      printf("\t *** Calibration values: ***\n\n");
      printf("  FREQ_OFFSET = %d Hz\n", eeconfig.s.freq_offset);
      printf("  R0_RSSI_OFFSET = %.1f dBm\n", eeconfig.s.radio0_rssi_offset);
      printf("  R1_RSSI_OFFSET = %.1f dBm\n", eeconfig.s.radio1_rssi_offset);
      printf("  SX1272_RSSI_OFFSET = %.1f dBm\n", eeconfig.s.sx1272_rssi_offset);

      /* LUT Gain Table */
      printf("\nLUT \tDIG \tPA \tMix \tRF_PWR\n");
      for(i=0; i<EEPROM_TX_LUT_MAX; i++)
      {
        printf("%d \t%u \t%u \t%u \t%d\n", i,
            eeconfig.s.tx_lut[i].dig_gain,
            eeconfig.s.tx_lut[i].pa_gain,
            eeconfig.s.tx_lut[i].mix_gain,
            eeconfig.s.tx_lut[i].rf_power);
      }

      printf("\n\nDUMP :\n");
      for(i = 0; i < (int)sizeof(eeconfig.raw); i++)
      {
        printf("%02X ",eeconfig.raw[i]);
        if((i%16) == 15) printf("\n");
      }
      printf("\n");
    }
	
    /* Close SPI */
    x = lgw_disconnect();
    if( x != 0 ) {
        printf( "ERROR: Failed to disconnect FPGA\n" );
        return EXIT_FAILURE;
    }

	return ret;
}

#else
int main(int argc, char **argv)
{
	printf("Not usable on this board\n");
	return -1;
}
#endif


/* --- EOF ------------------------------------------------------------------ */

