#include <errno.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef KERLINK_KEROS
#include <fslotpreg.h>
#endif

#include "loragw_fpga.h"
#include "loragw_reg.h"
#include "loragw_spi.h"

#include "parson.h"

#ifdef KERLINK
#include "fpga_lbt_863.h"
#include "fpga_lbt_915.h"
#endif
#include "fpga_spectral_scan.h"

#define info(f, a...)           fprintf(stdout, "I/ " f "\n", ##a)
#define warning(f, a...)        fprintf(stdout, "W/ " f "\n", ##a)
#define error(f, a...)          fprintf(stderr, "E/ " f "\n", ##a)
#define eerror(f, a...)         error(f ": %s", ##a, strerror(errno))

#if KERLINK_KEROS
#define CALIB_FILE              "/tmp/calib_rf.json"
#else
#define CALIB_FILE              "/mnt/fsuser-1/loraboard_conf.json"
#endif
#define INFO_FILE               "/tmp/loraboard_version"

static char *parse_eeprom_info(eeprom_config_t *config)
{
	uint32_t serial;
	char *type;
	char *buf;
	char unknownType[32];

	if ((buf = malloc(128)) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;

	}

	type = lgw_eeprom_get_type_name(config->s.type);
	if (type == NULL)
		snprintf(unknownType, sizeof(unknownType), "Type %u", config->s.type);

	serial = (config->s.serialno[0] << 16) +
	         (config->s.serialno[1] << 8) +
	         (config->s.serialno[2] << 0);

	snprintf(buf, 128,
	         "LORABOARD_MANUFACTURER=%.2X\n"
	         "LORABOARD_TYPE=\"%s\"\n"
	         "LORABOARD_HWVERSION=%.2X\n"
	         "LORABOARD_SERIALNO=%.6X\n",
	         config->s.manuf, type? type:unknownType,
	         config->s.hwver, serial);

	return buf;
}

#define json_set_or_die(type, root, name, value) { \
       if (json_object_dotset_##type(json_object(root), name, value) < 0) { \
               error("Failed to set json " #type ": %s", name); \
               json_value_free(root); \
               return NULL; \
       } \
}

static char *parse_eeprom_calib(eeprom_config_t *config)
{
	JSON_Value *root;
	char jpath[32];
	char *buf;
	uint8_t i;

	if ((root = json_value_init_object()) == NULL) {
		error("Failed to initialize json root");
		return NULL;
	}

	json_set_or_die(number, root, "SX1301_conf.radio_0.rssi_offset",
	                config->s.radio0_rssi_offset);
	json_set_or_die(number, root, "SX1301_conf.radio_1.rssi_offset",
	                config->s.radio1_rssi_offset);

	for (i = 0; i < 16; i++) {
		snprintf(jpath, sizeof(jpath), "SX1301_conf.tx_lut_%u.dig_gain", i);
		json_set_or_die(number, root, jpath, config->s.tx_lut[i].dig_gain);

		snprintf(jpath, sizeof(jpath), "SX1301_conf.tx_lut_%u.pa_gain", i);
		json_set_or_die(number, root, jpath, config->s.tx_lut[i].pa_gain);

		snprintf(jpath, sizeof(jpath), "SX1301_conf.tx_lut_%u.mix_gain", i);
		json_set_or_die(number, root, jpath, config->s.tx_lut[i].mix_gain);

		snprintf(jpath, sizeof(jpath), "SX1301_conf.tx_lut_%u.rf_power", i);
		json_set_or_die(number, root, jpath, config->s.tx_lut[i].rf_power);
	}

	if ((buf = json_serialize_to_string_pretty(root)) == NULL) {
		error("Failed to serialize json value");
		json_value_free(root);
		return NULL;
	}

	json_value_free(root);
	return buf;
}

static int default_calib(eeprom_config_t *config)
{
	eeprom_config_t default_868 = {
		.s = {
			.radio0_rssi_offset = -166,
			.radio1_rssi_offset = -166,
			.tx_lut = {
				[0] = { .dig_gain = 3, .pa_gain = 1, .mix_gain = 8, .rf_power = -2 },
				[1] = { .dig_gain = 3, .pa_gain = 1, .mix_gain = 9, .rf_power = 0 },
				[2] = { .dig_gain = 3, .pa_gain = 1, .mix_gain = 10, .rf_power = 1 },
				[3] = { .dig_gain = 3, .pa_gain = 1, .mix_gain = 11, .rf_power = 3 },
				[4] = { .dig_gain = 3, .pa_gain = 1, .mix_gain = 13, .rf_power = 7 },
				[5] = { .dig_gain = 3, .pa_gain = 2, .mix_gain = 9, .rf_power = 9 },
				[6] = { .dig_gain = 3, .pa_gain = 2, .mix_gain = 10, .rf_power = 11 },
				[7] = { .dig_gain = 3, .pa_gain = 2, .mix_gain = 11, .rf_power = 13 },
				[8] = { .dig_gain = 3, .pa_gain = 2, .mix_gain = 13, .rf_power = 16 },
				[9] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 9, .rf_power = 18 },
				[10] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 10, .rf_power = 20 },
				[11] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 11, .rf_power = 22 },
				[12] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 12, .rf_power = 24 },
				[13] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 13, .rf_power = 25 },
				[14] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 14, .rf_power = 26 },
				[15] = { .dig_gain = 3, .pa_gain = 3, .mix_gain = 15, .rf_power = 27 }
			}
		}
	};

	switch (config->s.type) {
	case 1:  /* 868-27dBm */
	case 16: /* WiFC-868 */
	case 0x56: /* WIIS-868 */
	case 0x64: /* FEVO-868 */
		memcpy(config, &default_868, sizeof(eeprom_config_t));
		return 0;

	default:
		error("No default calibration for this board");
		return -1;
	}
}

int main(int argc, char **argv)
{
	int opt, optidx;

	bool print_calib = false;
	bool print_info = false;
	bool program_fpga = true;
#ifdef KERLINK
	uint16_t lbt_freq = 0;
#endif

	uint8_t *fpga_data = fpga_spectral_scan;
	uint16_t fpga_size = sizeof(fpga_spectral_scan);
	char *fpga_name = "spectral scan";

	eeprom_config_t config;
	char *calib_str, *info_str;
	uint32_t *r0, *r1;
	FILE *file;

	const struct option lopt[] = {
#ifdef KERLINK
		{"lbt",      no_argument,        NULL, 'l' },
		{"freq",     required_argument,  NULL, 'f' },
#endif
		{"info",     no_argument,        NULL, 'i'},
		{"calib",    no_argument,        NULL, 'c'},
		{"version",  no_argument,        NULL, 'v'},
		{"help",     no_argument,        NULL, 'h'},
		{NULL,       0,                  NULL,  0 },
	};

	/* Manage options */
#ifdef KERLINK
#define OPTIONS                 "lf:icvh"
#else
#define OPTIONS                 "icvh"
#endif

	while ((opt = getopt_long(argc, argv, OPTIONS, lopt, &optidx)) != EOF) {
		switch (opt) {
#ifdef KERLINK

		case 'l':
			if (lbt_freq == 0) {
				lbt_freq = 915;
			}

			break;

		case 'f':
			switch ((lbt_freq = atoi(optarg))) {
			case 863:
			case 915:
				break;

			default:
				error("LBT firmware is not available "
				      "for this frequency");
				exit(EXIT_FAILURE);
			}

			break;
#endif

		case 'i':
			program_fpga = false;
			print_info = true;
			break;

		case 'c':
			program_fpga = false;
			print_calib = true;
			break;

		case 'v':
			fprintf(stdout, "util_fpga_start " PACKAGE_VERSION "\n");
			exit(EXIT_SUCCESS);

		default:
			fprintf(stderr,
			        "Usage: util_fpga_start [OPTIONS]...\n\n"
#ifdef KERLINK
			        "-l, --lbt                         "
			        "    Load LBT firmware.\n"
			        "-f, --freq FREQ_MHZ               "
			        "    Use LBT frequency (863 MHz or 915 MHz).\n"
			        "                                  "
			        "    Implies -l. Default to 915 MHz.\n"
#endif
			        "-i, --info                        "
			        "    Print board informations.\n"
			        "-c, --calib                       "
			        "    Print board calibration.\n"
			        "-v, --version                     "
			        "    Display version.\n"
			        "-h, --help                        "
			        "    Display this help screen.\n");
			exit(EXIT_FAILURE);
		}
	}

	/* Program FPGA */
	if (program_fpga == true) {
#ifdef KERLINK

		switch (lbt_freq) {
		case 863:
			fpga_name = "LBT 863 MHz";
			fpga_data = fpga_lbt_863;
			fpga_size = sizeof(fpga_lbt_863);
			break;

		case 915:
			fpga_name = "LBT 915 MHz";
			fpga_data = fpga_lbt_915;
			fpga_size = sizeof(fpga_lbt_915);
			break;
		}

#endif
		info("Programming FPGA with %s firmware", fpga_name);

		if (lgw_fpga_program(fpga_data, fpga_size) < 0) {
			error("Failed to program FPGA");
			exit(EXIT_FAILURE);
		}
	}

	/* Read EEPROM */
	if (lgw_connect(true, LGW_DEFAULT_NOTCH_FREQ) < 0) {
		error("Failed to connect FPGA");
		exit(EXIT_FAILURE);

	} else if (lgw_eeprom_readbuf(0, config.raw, sizeof(config.raw)) < 0) {
		error("Failed to read EEPROM");
		lgw_disconnect();
		exit(EXIT_FAILURE);
	}

	lgw_disconnect();

	if (print_info == true) {
		/* Print informations */
		if ((info_str = parse_eeprom_info(&config)) == NULL) {
			error("Failed to parse EEPROM informations");
			exit(EXIT_FAILURE);

		} else {
			fprintf(stdout, "%s", info_str);
			free(info_str);
		}

	} else if (print_calib == true) {
		/* Print calibration */
		if ((calib_str = parse_eeprom_calib(&config)) == NULL) {
			error("Failed to parse EEPROM calibration");
			exit(EXIT_FAILURE);

		} else {
			fprintf(stdout, "%s\n", calib_str);
			free(calib_str);
		}

	} else {
		/* Write informations */
		if (config.s.manuf == 0xFF) {
			info("EEPROM seems empty");
			error("Failed to write informations");

		} else if ((info_str = parse_eeprom_info(&config)) == NULL) {
			error("Failed to parse EEPROM informations");
			exit(EXIT_FAILURE);

		} else if ((file = fopen(INFO_FILE, "w")) == NULL) {
			eerror("Failed to open file: " INFO_FILE);
			free(info_str);
			exit(EXIT_FAILURE);

		} else {
			fprintf(file, "%s", info_str);
			free(info_str);
			fclose(file);
		}

		/* Generate default calibration */
		if (config.s.manuf != 0xFF) {
			r0 = (uint32_t *)&config.s.radio0_rssi_offset;
			r1 = (uint32_t *)&config.s.radio1_rssi_offset;

			if (*r0 == 0 || *r1 == 0 ||
			    *r0 == 0xFFFFFFFF || *r1 == 0xFFFFFFFF) {
				warning("Unable to read calibration");
				info("Generating default calibration");

				if (default_calib(&config) < 0) {
					error("Failed to set default calibration");
					exit(EXIT_FAILURE);
				}
			}
		}

		/* Write calibration */
		if ((calib_str = parse_eeprom_calib(&config)) == NULL) {
			error("Failed to parse EEPROM calibration");
			exit(EXIT_FAILURE);

		} else if ((file = fopen(CALIB_FILE, "w")) == NULL) {
			eerror("Failed to open file: " CALIB_FILE);
			free(calib_str);
			exit(EXIT_FAILURE);

		} else {
			fprintf(file, "%s\n", calib_str);
			free(calib_str);
			fclose(file);
		}
	}

	exit(EXIT_SUCCESS);
}
